# -*- coding: utf-8 -*-
"""Abstract EffectBase class that is used as a base for all other effects."""
from abc import ABC, abstractmethod


class EffectBase(ABC):
    """
    Base effect class.
    
    This class is a part of generalized random effects model. It provides
    interface for ModelGeneralizedEffects to infer sample-wise covariance
    structure.
    """
    def __init__(self, columns=None, d_mode='diag'):
        """
        Instantiate EffectBase abstract class.

        Parameters
        ----------
        columns : tuple, str
            Column names that contain information on the data structure that
            is present in the dataset (e.g. time, region, etc).
        d_mode : str
            Mode of D matrix. If "diag", then D has unique params on the
            diagonal. If "full", then D is fully parametrised. If
            "scale", then D is an identity matrix, multiplied by a single
            variance parameter (scalar). The default is "diag".

        Returns
        -------
        None.

        """
        if type(columns) is str:
            columns = (columns, )
        elif columns is None:
            columns = list()
        self.columns = columns
        self.d_mode = d_mode
        self.parameters = None

    @abstractmethod
    def load(self, order, model, data, clean_start=True, **kwargs):
        """
        Called by model new dataset is loaded.
        
        Here, Effects are configured from the data. self.parameters must be
        initialised after invoking this method.
        Parameters
        ----------
        order : int
            Identificator of effect in model. It is just an order of the effect
            among other effects as specified by user.
        model : ModelGeneralizedEffects
            Instance of ModelGeneralizedEffects that calls this method.
        data : pd.DataFrame
            Dataset that is being loaded. Should contain self.columns.
        clean_start : bool, optional
            If True, then parameters are (re)initialized. The model will use
            the ones already present in self.parameters vector otherwise. The
            default is True.

        Returns
        -------
        None.

        """
        for col in self.columns:
            if col not in data.columns:
                raise Exception(f"Effect missing necessary {col} column in"
                                " data.")
        self.order = order

    @abstractmethod
    def calc_k(self, model):
        """
        Build numpy matrix that is used as a skeleton for the covariance
        across samples matrix K.
        
        Parameters
        -------
        model : ModelGeneralizedEffects
            Instance of ModelGeneralizedEffects that calls this method.
        Returns
        -------
        np.ndarray
            Matrix.
        tuple
            Tuple of rownames and colnames.

        """
        pass

    @abstractmethod
    def calc_k_grad(self, model):
        """
        Calculate gradient for matrix D.

        Parameters
        ----------
        model : ModelGeneralizedEffects
            Instance of ModelGeneralizedEffects that calls this method.
        Returns
        -------
        List[np.ndarray] of derivatives of K wrt to parameters.

        """
        pass

    def get_bounds(self):
        """
        Return bounding intervals for each of the effects parameters.

        Returns
        -------
        List[tuple]
            List of parameters intervals.

        """
        return [(None, None)] * len(self.parameters)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import unittest
import numpy as np
import pandas as pd
from ..model import Model
from ..examples import univariate_regression, multivariate_regression

np.random.seed(2021)
n = 100
p = 3
params = [np.random.uniform(0.2, 1.2, size=(p - 1, 1)),
          np.random.uniform(0.2, 1.2, size=(p -1, 1))]
params = list(map(lambda x: np.append([1], x), params))
y = np.random.normal(size=(n, 2 * p))
eta1 = np.random.normal(scale=1, size=(n, 1))
eta2 = np.random.normal(scale=1, size=(n, 1)) + 3 * eta1
y[:, :p] += np.kron(params[0], eta1)
y[:, p:] += np.kron(params[1], eta2)

res = list()
d = {'eta1': list(), 'eta2': list()}
y_names = list()
for i in range(1, p + 1):
    res.append((f'y{i}', '~', 'eta1', params[0][i - 1]))
    y_names.append(res[-1][0])
    d['eta1'].append(y_names[-1])
for j in range(1, p + 1):
    res.append((f'y{j + i}', '~', 'eta2', params[1][j - 1]))
    y_names.append(res[-1][0])
    d['eta2'].append(y_names[-1])
desc = '\n'.join(f"{eta} =~ {' + '.join(ys)}" for eta, ys in d.items())
desc += '\neta2 ~ eta1'
params = pd.DataFrame.from_records(res, columns=['lval', 'op', 'rval',
                                                 'Estimate'])
data = pd.DataFrame(np.append(np.append(y, eta1, axis=1), eta2, axis=1),
                    columns=y_names + ['eta1', 'eta2'])


class TestModel(unittest.TestCase):
    def evaluate(self, desc: str, data: pd.DataFrame, true: pd.DataFrame,
                 obj='MLW'):
        m = Model(desc)
        r = m.fit(data, obj=obj)
        assert r.success, f"Optimization routine failed. [{obj}]"
        ins = m.inspect()
        errs = list()
        for _, row in true.iterrows():
            t = (ins['op'] == row['op']) & (ins['lval'] == row['lval']) &\
                (ins['rval'] == row['rval'])
            if sum(t) == 0:
                continue
            t = ins[t]
            try:
                assert t['p-value'].values[0] < 0.05,\
                       f"Incorrect p-value estimate [{obj}]."
            except TypeError:
                pass
            est = t['Estimate'].values[0]
            errs.append(abs((est - row['Estimate']) / row['Estimate']))
        err = np.mean(errs)
        assert err < 0.1, \
               f"Parameter estimation quality is too low: {err} [{obj}]"
    
    def test_univariate_regression(self):
        desc = univariate_regression.get_model()
        data = univariate_regression.get_data()
        true = univariate_regression.get_params()
        self.evaluate(desc, data, true, 'MLW')
        self.evaluate(desc, data, true, 'ULS')
        self.evaluate(desc, data, true, 'GLS')
        self.evaluate(desc, data - data.mean(), true, 'FIML')

    def test_multivariate_regression(self):
        desc = multivariate_regression.get_model()
        data = multivariate_regression.get_data()
        true = multivariate_regression.get_params()
        self.evaluate(desc, data, true, 'MLW')
        self.evaluate(desc, data, true, 'ULS')
        self.evaluate(desc, data, true, 'GLS')
        self.evaluate(desc, data - data.mean(), true, 'FIML')

    def test_random_model(self):
        global params
        global desc
        global data
        self.evaluate(desc, data, params, 'MLW')
        self.evaluate(desc, data, params, 'ULS')
        self.evaluate(desc, data, params, 'GLS')
        self.evaluate(desc, data - data.mean(), params, 'FIML')



# -*- coding: utf-8 -*-
"""Effect with a pre-defied K matrix as in ModelEffects."""
from ..utils import calc_zkz
from .effect_base import EffectBase
import pandas as pd
import numpy as np


class EffectStatic(EffectBase):
    """
    Effect with a pre-defined static K matrix as in classical LMMs.
    
    This effect introduces no extra parameters other than those that come along
    the D matrix.
    """
    def __init__(self, columns: str, k: pd.DataFrame, d_mode='diag'):
        """
        Instantiate EffectStatic.

        Parameters
        ----------
        columns : str
            Name of column that corresponds to individuals group id. Should
            match the appropriate row/column in the K dataframe.
        k : pd.DataFrame
            K matrix with columns and rows subscribed with accorance to the
            existing "groups" that encapsulate individual observations.
        d_mode : str
            Mode of D matrix. If "diag", then D has unique params on the
            diagonal. If "full", then D is fully parametrised. If
            "scale", then D is an identity matrix, multiplied by a single
            variance parameter (scalar). The default is "diag".

        Returns
        -------
        None.

        """
        
        super().__init__(columns, d_mode=d_mode)
        self.k = k

    def load(self, i, model, data, clean_start=True, **kwargs):
        """
        Called by model new dataset is loaded.
        
        Here, Effects are configured from the data. self.parameters must be
        initialised after invoking this method.
        Parameters
        ----------
        order : int
            Identificator of effect in model. It is just an order of the effect
            among other effects as specified by user.
        model : ModelGeneralizedEffects
            Instance of ModelGeneralizedEffects that calls this method.
        data : pd.DataFrame
            Dataset that is being loaded. Should contain self.columns.
        clean_start : bool, optional
            If True, then parameters are (re)initialized. The model will use
            the ones already present in self.parameters vector otherwise. The
            default is True.

        Returns
        -------
        None.

        """
        super().load(i, model, data, clean_start, **kwargs)
        if clean_start:
            self.parameters = np.array([])
        mx_k = kwargs.get(f'k_{self.order+1}', self.k)
        c = data[self.columns[0]]
        self.mx_k = calc_zkz(c, mx_k)

    def calc_k(self, model):
        return self.mx_k

    def calc_k_grad(self, model):
        return []


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Explarotary Factor Analysis (EFA).

Methods that help to retrieve latent structure CFA-style. Here only our own
unorthodox approach is considered: we determine number of latent factors and
their loadings via clustering analysis.
"""
import numpy as np
import pandas as pd
from .model import Model
from collections import defaultdict
from sklearn.cluster import OPTICS
from sklearn.decomposition import SparsePCA
from itertools import permutations
from copy import deepcopy
from .utils import cor


def find_latents(data: pd.DataFrame, min_loadings=2, mx_cor=None,
                 underscript='', mode='spca', spca_alpha=1):
    """
    Retrieve number of latent factors and their simple loadings.

    Simple loadings here mean the simplest model where none of latent factors
    have joint indicator. Furthermore, this methods performs no SEM-based
    verification of the proposed CFA-model.
    Parameters
    ----------
    data : pd.DataFrame
        Data.
    min_loadings : int, optional
        Minimal number of indicators per latent factor. The default is 2.
    mx_cor : pd.DataFrame, optional
        Correlation matrix will be used instead of data if provided. The
        default is None.
    underscript : str, optional
        Underscript to add after latent factor names. The default is ''.
    mode : str, optional
        If 'optics', then a heuristics appraoch based entirely on OPTICS is
        used. If "spca", then OPTICS is used only to find number of latent
        factors, and loadings are found using sprase PCA. The default is 
        'optics'.
    spca_alpha : float, optional
        If mode == 'spca', then it is a regularization multiplier. The default
        is 1.0.
    Returns
    -------
    tuple
        Mapping latent factor->indicators and distance matrix.

    """
    if mx_cor is None:
        names = data.columns
        mx_cor = cor(data)
    else:
        names = mx_cor.columns
        mx_cor = mx_cor.values
    names = list(names)
    dist = np.clip(1.0 - np.abs(mx_cor), 0.0, 1.0)
    clust = OPTICS(min_samples=min_loadings, metric='precomputed').fit(dist)
    loadings = defaultdict(set)
    if mode == 'optics':
        for i, label in enumerate(clust.labels_):
            if label != -1:
                name = f'eta{label+1}{underscript}'
            else:
                name = -1
            loadings[name].add(names[i])
    else:
        num_lats = set(clust.labels_)
        num_lats = len(num_lats - {-1})
        if data is None:
            raise Exception("For sparse PCA mode data must be provided.")
        cmp = SparsePCA(num_lats, alpha=spca_alpha).fit(data).components_
        for i in range(cmp.shape[0]):
            inds = np.where(abs(cmp[i]) >= 0.05)[0]
            inds = data.columns[inds]
            name = f'eta{i+1}{underscript}'
            loadings[name] = set(inds)
    for lat, inds in loadings.items():
        if lat == -1:
            continue
        dists = list()
        inds = list(inds)
        for ind in inds:
            d = 0
            i = names.index(ind)
            for ind0 in inds:
                if ind0 != ind:
                    j = names.index(ind0)
                    d += dist[i, j]
            dists.append(d)
        inds = sorted(inds, key=lambda x: dists[inds.index(x)])
        loadings[lat] = inds
    return loadings, pd.DataFrame(dist, columns=names, index=names)


def dict_to_desc(d: dict):
    """
    Transform dictionary into a text-based description.

    Parameters
    ----------
    d : dict
        Mapping factors->indicators.

    Returns
    -------
    desc : str
        Text-form description of the model.

    """
    desc = str()
    for lat, inds in d.items():
        if lat != -1:
            inds = ' + '.join(inds)
            desc += f'{lat} =~ {inds}\n'
    return desc


def finalize_loadings(loadings: dict, data: pd.DataFrame, dist: pd.DataFrame,
                      pval=0.01, model=Model, base_desc='', only_clean=False):
    """
    Test p-values of CFA-like model and find cross-loadings between factors.

    Parameters
    ----------
    loadings : dict
        Mapping factor->indicators.
    data : pd.DataFrame
        Dataset.
    dist : pd.DataFrame
        Distance matrix.
    pval : float, optional
        Statistical significiance cut-off value. The default is 0.01.
    model : TYPE, optional
        Class instance. The default is Model.
    base_desc : str, optional
        Text description to append to each model. The default is ''.
    only_clean : bool, optional
        If True, then only high p-value loadings are dropped, no
        intercorrelations are studied. The default is False.

    Returns
    -------
    dict
        Mapping factor->indicators.

    """

    def clean_loadings(loadings: dict, base_desc: str):
        desc = dict_to_desc(loadings)
        if base_desc:
            desc += '# Base desc:\n' + base_desc
        m = model(desc)
        m.fit(data)
        ins = m.inspect()
        lats_to_remove = set()
        for lat, inds in loadings.items():
            if lat != -1:
                inds = iter(inds)
                # if not base_desc:
                next(inds)
                to_remove = set()
                for ind in inds:
                    if not is_loading_significant(ins, ind, lat, pval=pval):
                        to_remove.add(ind)
                        if -1 in loadings:
                            loadings[-1].add(ind)
                it = loadings[lat]
                list(map(lambda x: it.remove(x), to_remove))
                if len(it) < 2:
                    lats_to_remove.add(lat)
        for lat in lats_to_remove:
            del loadings[lat]
        if -1 in loadings:
            for lat, inds in list(loadings.items()):
                if len(inds) == 1 and lat != -1:
                    del loadings[lat]
                    loadings[-1].append(inds[0])

    def test(loadings: dict, base_desc: str, lat: str, ind: str):
        desc = dict_to_desc(loadings)
        desc += base_desc
        desc += f'{lat} =~ {ind}'
        m = model(desc)
        if not m.fit(data).success:
            return float('inf')
        ins = m.inspect()
        if m._fim_warn:
            return float('inf')
        return get_loading_significiance(ins, ind, lat)
    loadings = deepcopy(loadings)
    clean_loadings(loadings, base_desc)
    if only_clean:
        return loadings
    loadings_comp = deepcopy(loadings)
    try:
        del loadings_comp[-1]
    except KeyError:
        pass
    lats = [lat for lat in loadings if lat != -1]
    for ind in loadings[-1]:
        pvals = [test(loadings, base_desc, lat, ind) for lat in lats]
        i = np.argmin(pvals)
        if pvals[i] < pval:
            loadings_comp[lats[i]].append(ind)
    loadings = loadings_comp
    clean_loadings(loadings, base_desc)
    loadings_joint = deepcopy(loadings)
    for a, b in permutations(loadings, 2):
        b_items = loadings[b]
        a_items = sorted(loadings[a],
                         key=lambda x: min(dist.loc[x, b_items]))
        for item in a_items:
            kt = test(loadings_joint, base_desc, b, item)
            if kt < pval:
                loadings_joint[b].append(item)
            else:
                # continue
                break
    tmp = None
    while loadings_joint != tmp:
        tmp = deepcopy(loadings_joint)
        clean_loadings(loadings_joint, base_desc)
    return loadings_joint


def explore_cfa_model(data: pd.DataFrame, min_loadings=2, pval=0.01,
                      model=Model, ret_desc=True, mode='spca',
                      spca_alpha=1.0):
    """
    Retrieve CFA model from data.

    Parameters
    ----------
    data : pd.DataFrame
        Dataset.
    min_loadings : int, optional
        Minimal number of indicators per factor. The default is 2.
    pval : TYPE, optional
        P-value cutoff. The default is 0.01.
    model : TYPE, optional
        Class to instantiate. The default is Model.
    ret_desc : TYPE, optional
        If True, text-based description is returned. If False, a mapping is
        returned instead. The default is True.
    mode : str, optional
        If 'optics', then a heuristics appraoch based entirely on OPTICS is
        used. If "spca", then OPTICS is used only to find number of latent
        factors, and loadings are found using sprase PCA. The default is 
        'optics'.
    spca_alpha : float, optional
        If mode == 'spca', then it is a regularization multiplier. The default
        is 1.0.

    Returns
    -------
    str
        Model description.

    """
    loadings, dist = find_latents(data, min_loadings=min_loadings, mode=mode,
                                  spca_alpha=spca_alpha)
    loadings = finalize_loadings(loadings, data=data, dist=dist, pval=pval,
                                 model=Model, only_clean=mode == 'spca')
    return dict_to_desc(loadings) if ret_desc else loadings


def explore_pine_model(data: pd.DataFrame, min_loadings=2, pval=0.01, levels=2,
                       model=Model):
    """
    Retrieve pine-like model from the data.

    "Pine-like" means that we also assume a layer of latent factors above the
    CFA model that regresses onto latent factors in the CFA model.
    Parameters
    ----------
    data : pd.DataFrame
        Dataset.
    min_loadings : int, optional
        Minimal number of loadings per factor. The default is 2.
    pval : float, optional
        P-value cutoff. The default is 0.01.
    levels : int, optional
        Maximal number of layers. The default is 2.
    model : Model, optional
        Class to instantiate. The default is Model.

    Returns
    -------
    str
        Model description.

    """
    cfa = explore_cfa_model(data, min_loadings=min_loadings, pval=pval,
                            model=model, ret_desc=False)
    to_remove = {lat for lat, items in cfa.items() if len(items) < 2}
    for lat in to_remove:
        del cfa[lat]
    pine = dict_to_desc(cfa)
    names = list(cfa.keys())
    min_loadings = 2
    for level in range(1, levels):
        m = model(pine)
        m.fit(data)
        psi = m.inspect('mx')['Psi'].loc[names, names]
        d = np.diag(psi.values.diagonal() ** (-0.5))
        corr = d @ psi @ d
        corr.columns = names
        corr.index = names
        lats, dist = find_latents(None, mx_cor=corr, min_loadings=2,
                                  underscript=f'_{level}')
        joint = finalize_loadings(lats, data, dist, pval=pval, model=model,
                                  base_desc=pine)
        if joint:
            pine += 'DEFINE(latent) {}\n'.format(' '.join(joint.keys()))
        else:
            break
        names = list(joint.keys())
        pine += dict_to_desc(joint)
    return pine.strip()


def get_loading_significiance(res, lval: str, rval: str):
    """
    Get p-value of certain regression coefficient.

    Parameters
    ----------
    res : Model or pd.DataFrame
        Either fitted Model or pandas DataFrame.
    lval : str
        Left-value of operation.
    rval : str
        Right-value of operation.

    Returns
    -------
    float
        p-value.

    """
    if type(res) is not pd.DataFrame:
        res = res.inspect()
    inds = (res.lval == lval) & (res.rval == rval) & (res.op == '~')
    p = res[inds]['p-value'].values[0]
    return p


def is_loading_significant(res, lval: str, rval: str, pval=0.01):
    """
    Test significiance of a regression coefficient.

    Parameters
    ----------
    res : Model or pd.DataFrame
        Either fitted Model or pandas DataFrame.
    lval : str
        Left-value of operation.
    rval : str
        Right-value of operation.
    pval : TYPE, optional
        p-value cutoff. The default is 0.01.

    Returns
    -------
    bool
        True if significiant, False otherwise.
    """
    p = get_loading_significiance(res, lval, rval)
    if p < pval:
        return True
    return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Model example from semopy article."""
import pandas as pd
import numpy as np
import os

__desc = '''# Measurement part
eta1 =~ y1 + y2 + y3
eta2 =~ y3 + y2
eta3 =~ y4 + y5
eta4 =~ y4 + y6
# Structural part
eta3 ~ x2 + x1
eta4 ~ x3
x3 ~ eta1 + eta2 + x1
x4 ~ eta4 + x6
y7 ~ x4 + x6
# Additional covariances
y6 ~~ y5
x2 ~~ eta2'''


__folder = os.path.dirname(os.path.abspath(__file__))
__filename = '%s/article_data.csv' % __folder
__u_filename = '%s/article_data_u.npy' % __folder
__u2_filename = '%s/article_data_u2.npy' % __folder
__k_filename = '%s/article_data_k.npy' % __folder
__k2_filename = '%s/article_data_k2.npy' % __folder
__v_filename = '%s/article_data_u_vars.txt' % __folder
__ma_filename = '%s/article_data_ma.npy' % __folder
__mv_filename = '%s/article_data_ma_vars.txt' % __folder
__params_filename = '%s/article_params.csv' % __folder

def get_model():
    """
    Retrieve model description in semopy syntax.

    Returns
    -------
    str
        Model's description.

    """
    return __desc


def get_data(drop_factors=True, random_effects=0, moving_average=False):
    """
    Retrieve dataset.
    
    Parameters
    -------
    drop_factors : bool, optional
        If True, then factors are dropped from the dataframe. The default is
        True.
    random_effects : int, optional
        Can be 0, 1 or 2: number of random effects that "spoil" the data.
        If non-zero, then data contaminated with random effects together with
        covariance matrix K is returned instead.
    moving_average : bool, optional
        If True, then data is also 'contaminated' with MA(2) noise. The default
        is False.

    Returns
    -------
    pd.DataFrame
        Dataset.

    """
    data = pd.read_csv(__filename, sep=',', index_col=0)
    if drop_factors:
        etas = [v for v in data.columns if v.startswith('eta')]
        data = data.drop(etas, axis=1)
    if moving_average:
        data['time'] = list(map(float, data.index))
        ma = np.load(__ma_filename)
        with open(__mv_filename, 'r') as f:
            cols = f.read().split(' ')
        data[cols] += ma
    if random_effects:
        data['group'] = data.index
        k = np.load(__k_filename)
        k = pd.DataFrame(k, index=data['group'], columns=data['group'])
        u = np.load(__u_filename)
        with open(__v_filename, 'r') as f:
            cols = f.read().split(' ')
        data[cols] += u
        if random_effects == 1:
            return data, k
        k2 = np.load(__k2_filename)
        k2 = pd.DataFrame(k2, index=data['group'], columns=data['group'])
        u = np.load(__u2_filename)
        data[cols] += u
        return data, (k, k2)
    return data

def get_params():
    """
    Retrieve true parameter values.

    Returns
    -------
    None.

    """
    return pd.read_csv(__params_filename, sep=',', index_col=0)
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Political Democracy model and data.

Bollen's Data on Industrialization and Political Democracy.
Dataset set includes four measures of democracy at two points in time, 1960
and 1965, and three measures of industrialization in 1960, for 75 developing
countries.
"""
import pandas as pd
import os

__desc = '''# measurement model
ind60 =~ x1 + x2 + x3
dem60 =~ y1 + y2 + y3 + y4
dem65 =~ y5 + y6 + y7 + y8
# regressions
dem60 ~ ind60
dem65 ~ ind60 + dem60
# residual correlations
y1 ~~ y5
y2 ~~ y4 + y6
y3 ~~ y7
y4 ~~ y8
y6 ~~ y8'''

__filename = '%s/pd_data.csv' % os.path.dirname(os.path.abspath(__file__))


def get_model():
    """
    Retrieve model description in semopy syntax.

    Returns
    -------
    str
        Model's description.

    """
    return __desc


def get_data():
    """
    Retrieve dataset.

    Returns
    -------
    pd.DataFrame
        Dataset.

    """
    return pd.read_csv(__filename, index_col=0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Base model module that contains ModelBase for lienear SEM models."""
from abc import ABC, abstractmethod
from collections import defaultdict
from .parser import parse_desc


class ModelBase(ABC):
    """
    Base model class.

    Model is a base class for linear semopy SEM models. It bears the burden of
    instantiating base SEM relatioships for their further explotation
    in practical semopy SEM models.
    """

    symb_regression = '~'
    symb_covariance = '~~'
    symb_measurement = '=~'
    symb_force_variables = 'FORCE'
    symb_define = 'DEFINE'
    set_types = set()
    dict_effects = dict()
    dict_operations = dict()

    def __init__(self, description: str):
        """
        Instantiate base model.

        Model is a base class for linear semopy SEM models. It bears the
        of burdens instantiating base SEM relatioships for their further
        exploitation in practical semopy SEM models.
        Parameters
        ----------
        description : str
            Model description in semopy syntax.

        Returns
        -------
        None.

        """
        self.dict_effects[self.symb_regression] = self.effect_regression
        self.dict_effects[self.symb_covariance] = self.effect_covariance
        self.dict_effects[self.symb_measurement] = self.effect_measurement
        self.dict_operations[self.symb_force_variables] = self.operation_force
        self.dict_operations[self.symb_define] = self.operation_define
        self.description = description
        if type(description) is str:
            effects, operations = parse_desc(description)
        else:
            effects, operations = description
        self.before_classification(effects, operations)
        self.classify_variables(effects, operations)
        self.post_classification(effects)
        self.create_parameters(effects)
        self.apply_operations(operations)
        self.finalize_init()

    def before_classification(self, effects: dict, operations: dict):
        """
        Preprocess effects and operations if necessary before classification.

        Parameters
        ----------
        effects : dict
            Dict returned from parse_desc.

        operations: dict
            Dict of operations as returned from parse_desc.

        Returns
        -------
        None.

        """
        pass

    def classify_variables(self, effects: dict, operations: dict):
        """
        Classify and instantiate vars dict.

        Parameters
        ----------
        effects : dict
            Dict returned from parse_desc.

        operations: dict
            Dict of operations as returned from parse_desc.

        Returns
        -------
        None.

        """
        self.vars = dict()
        latents = set()
        in_arrows, out_arrows = set(), set()
        senders = defaultdict(set)
        indicators = set()
        for rv, inds in effects[self.symb_measurement].items():
            latents.add(rv)
            indicators.update(inds)
            out_arrows.add(rv)
            senders[rv].update(inds)
        in_arrows.update(indicators)
        for rv, lvs in effects[self.symb_regression].items():
            in_arrows.add(rv)
            out_arrows.update(lvs)
            for lv in lvs:
                senders[lv].add(rv)
        for operation in operations[self.symb_define]:
            if operation.params and operation.params[0] == 'latent':
                latents.update(operation.onto)
        for operation in operations[self.symb_force_variables]:
            if operation.params and operation.params[0] == 'endo':
                in_arrows.update(operation.onto)
        allvars = out_arrows | in_arrows
        exogenous = out_arrows - in_arrows
        outputs = in_arrows - out_arrows
        endogenous = allvars - exogenous
        observed = allvars - latents
        self.vars['all'] = allvars
        self.vars['endogenous'] = endogenous
        self.vars['exogenous'] = exogenous
        self.vars['observed'] = observed
        self.vars['latent'] = latents
        self.vars['indicator'] = indicators
        self.vars['output'] = outputs
        self.vars_senders = senders

    def create_parameters(self, effects: dict):
        """
        Instantiate parameters in a model.

        Parameters
        ----------
        effects : dict
            Mapping of effects as returned by parse_desc.

        Raises
        ------
        NotImplementedError
            Raises in case of unknown effect symbol.

        Returns
        -------
        None.

        """
        for operation, items in effects.items():
            try:
                self.dict_effects[operation](items)
            except KeyError:
                raise NotImplementedError(f'{operation} is an unknown op.')

    def apply_operations(self, operations: dict):
        """
        Apply operations to model.

        Parameters
        ----------
        operations : dict
            Mapping of operations as returned by parse_desc.

        Raises
        ------
        NotImplementedError
            Raises in case of unknown command name.

        Returns
        -------
        None.

        """
        for command, items in operations.items():
            try:
                list(map(self.dict_operations[command], items))
            except KeyError:
                raise NotImplementedError(f'{command} is an unknown command.')

    def post_classification(self, effects: dict):
        """
        Procedure that is run just after classify_variables.

        Parameters
        -------
        effects : dict
            Maping opcode->values->rvalues->mutiplicator.

        Returns
        -------
        None.

        """
        pass

    def operation_force(self, operation):
        """
        Works through FORCE command.

        Parameters
        ----------
        operation : Operation
            Operation namedtuple.

        Returns
        -------
        None.

        """
        pass

    def operation_define(self, operation):
        """
        Works through DEFINE command.

        Parameters
        ----------
        operation : Operation
            Operation namedtuple.

        Returns
        -------
        None.

        """
        pass

    @abstractmethod
    def effect_regression(self, items):
        """
        Works through regression operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        pass

    @abstractmethod
    def effect_measurement(self, items):
        """
        Works through measurement operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        pass

    @abstractmethod
    def effect_covariance(self, items):
        """
        Works through covariance operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        pass

    @abstractmethod
    def fit(self, **kwargs):
        """
        Fits model to data.

        Returns
        -------
        None.

        """
        pass

    def finalize_init(self):
        """
        Run a post-command processing routine in the end of initialization.

        Returns
        -------
        None.

        """
        pass



# -*- coding: utf-8 -*-
"""Second order de-biasing correction"""
from .model_generation import generate_data
from copy import deepcopy
import pandas as pd
import numpy as np

def bias_correction(model, n=100, resample_mean=False, extra_data=None,
                    max_rel_fun=3, clean_slate=False, **kwargs):
    """
    Parametric Bootstrap bias correction.

    Parameters
    ----------
    model : TYPE
        Model or ModelMeans. Tehnically, ModelEffects and
        ModelGeneralizedEffects will work too, but the results are almost
        guaranteed to be bad. It will probably change eventually, but not
        anytime soon.
    n : int, optional
        Number of bootstrap iterations. The default is 100.
    resample_mean : bool, optional
        If True, then mean components in ModelMeans are also resampled. The
        default is False.
    extra_data : pd.DataFrame, optional
        Extra dataframe to append to resampled data. The default is None.
    max_rel_fun : float, optional
        Bootstrap iteration is skipped if optimizer failed or if the resulting
        loss function is greater than max_rel_fun times original MLE loss fun.
        The default is 3.
    clean_slate : bool, optional
        If True, then each PBS iteration uses previous values as starting ones.
        May speed-up computations. The default is True.
    **kwargs : dict
        Extra arguments to be passed to fit method.

    Returns
    -------
    None.

    """
    
    model_c = deepcopy(model)
    t = None
    try:
        n_samples = model.n_samples
    except AttributeError:
        n_samples = model.mx_data[0]
    if extra_data is not None:
        extra_cols = set(extra_data.columns)
    i = 0
    row_fails = 0
    px = model.param_vals.copy()
    fun = model.last_result.fun
    while i < n:
        if not resample_mean:
            data = generate_data(model, n=n_samples)
        else:
            data = generate_data(model, n=n_samples, generator_exo=None)
        if extra_data is not None:
            data.index = extra_data.index
            st = extra_cols - set(data.columns)
            data = pd.concat([data, extra_data[st]], axis=1)
        r = model_c.fit(data, clean_slate=clean_slate, **kwargs)
        if not r.success or (r.fun > max_rel_fun * fun):
            row_fails += 1
            if row_fails == 20:
                raise np.linalg.LinAlgError("Couldn't sample proper data.")
            continue
        if t is None:
            t = r.x
        else:
            t += r.x
        model.param_vals[:] = px
        i += 1
        row_fails = 0
    model.param_vals = 2 * px - t / n
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Multivariate (3) regression with 3 independent variables example model."""
import pandas as pd
import os

__desc = 'y1, y2, y3 ~ x1 + x2 + x3'

__folder = os.path.dirname(os.path.abspath(__file__))
__filename = '%s/multivariate_regression_data.csv' % __folder
__filename_p = '%s/multivariate_regression_params.csv' % __folder


def get_model():
    """
    Retrieve model description in semopy syntax.

    Returnsunivariate_regression
    -------
    str
        Model's description.

    """
    return __desc


def get_data():
    """
    Retrieve dataset.

    Returns
    -------
    pd.DataFrame
        Dataset.

    """
    return pd.read_csv(__filename, index_col=0)

def get_params():
    """
    Retrieve true parameter estimates.

    Returns
    -------
    pd.DataFrame
        Dataset.

    """
    return pd.read_csv(__filename_p, index_col=0)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Simple univariate BLUP implementation for starting values estimation."""
import numpy as np
from scipy.optimize import minimize


def grad(sigmas: np.ndarray, y: np.ndarray, k: np.ndarray):
    v = 1 / (sigmas[0] + sigmas[1] * k)
    if np.any(v < 1e-8):
        return [np.nan, np.nan]
    yt = y * v
    g = np.zeros(2)
    g[0] = np.sum(yt ** 2) - np.sum(np.log(v))
    g[1] = np.sum(yt * k * y) - np.sum(np.log(v ** 2 * k))
    return g

def obj(sigmas: np.ndarray, y: np.ndarray, k: np.ndarray):
    v = 1 / (sigmas[0] + sigmas[1] * k)
    if np.any(v < 1e-8):
        return np.nan
    yt = y * v
    return np.sum(yt * y) - np.sum(np.log(v))

def blup(y: np.ndarray, k: np.ndarray, p=0.8, maxiter=50):
    """
    Calculate BLUP estimate for U of a single variable.

    Parameters
    ----------
    y : np.ndarray
        Observations of a given variable.
    k : np.ndarray
        K matrix.
    p : float, optional
        Expected ratio of variable variance to random effect variance. Used for
        starting values only. The default is 0.8.
    maxiter : int, optional
        Maximal number of iterations. Better not be too high or estimation
        process could take noticable time in some cases. The default is 50.

    Returns
    -------
    U
        Random effects estimate (BLUP).

    """
    
    v = np.var(y)
    x0 = np.array([p * v, (1- p) * v])
    s = minimize(lambda x: obj(x, y, k), x0, jac=lambda x: grad(x, y, k),
                 method="SLSQP", options={'maxiter': maxiter},
                 bounds=([0, None], [0, None])
                 ).x
    v = 1/(1 / s[0] + (1 / s[1]) * (1/k))
    return y * v / s[0], s



from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(	
      install_requires=['scipy', 'numpy', 'pandas', 'sympy', 'sklearn', 
                       		    'statsmodels'],
      include_package_data=True,
      package_data={'': ['examples/*.csv', 'examples/*.npy', 'report/*.html', 'report/*.txt',
					'examples/*.txt', 'report/css/*.css', 'report/js/*.js']},
      name="semopy",
      version="2.3.0",
      author="Georgy Meshcheryakov",
      author_email="metsheryakov_ga@spbstu.ru",
      description="Structural Equation Modeling Optimization in Python.",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://semopy.com",
      packages=find_packages(),
      python_requires=">=3.7",
      classifiers=[
              "Programming Language :: Python :: 3.7",
	      "Programming Language :: Python :: 3.8",
	      "Programming Language :: Python :: 3.9",
	      "Programming Language :: Python :: 3.10",
              "License :: OSI Approved :: MIT License",
	      "Development Status :: 5 - Production/Stable",
	      "Topic :: Scientific/Engineering",
              "Operating System :: OS Independent"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""An example of dataset with random effects.

Correct parameter estimates:
       lval  op  rval     Value
    0    x2   ~  eta1  2.008729
    1  eta2   ~  eta1  2.879663
    2    x1   ~  eta1 -2.292444
    3  eta1  =~    y2  2.974733
    4  eta2  =~    y4 -1.246382
Notice that get_data() returns tuple: dataset and K matrix.
"""
import pandas as pd
import os

__desc = '''eta1 =~ y1 + y2
eta2 =~ y3 + y4
x2 ~ eta1
eta2 ~ eta1
x1 ~ eta1'''

__folder = os.path.dirname(os.path.abspath(__file__))
__filename = '%s/example_rf_data.csv' % __folder
__filename_k = '%s/example_rf_kinship.csv' % __folder


def get_model():
    """
    Retrieve model description in semopy syntax.

    Returns
    -------
    str
        Model's description.

    """
    return __desc


def get_data():
    """
    Retrieve dataset and kinship matrix.

    Returns
    -------
    pd.DataFrame
        Dataset and K matrix.

    """
    data = pd.read_csv(__filename, index_col=0)
    data['group'] = data.index
    k = pd.read_csv(__filename_k, index_col=0)
    k.columns = list(map(int, k.columns))
    return data, k



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Internal imputer module."""

from .model import Model
from .model_means import ModelMeans
from .model_effects import ModelEffects
from .solver import Solver
from .utils import chol_inv
from . import startingvalues
import pandas as pd
import numpy as np


class Imputer(Model):
    """Model for missing data imputation."""

    symb_missing = '@'

    matrices_names = tuple(list(Model.matrices_names) + ['data_imp'])

    def __init__(self, model: Model, data: pd.DataFrame, factors=True):
        """
        Instantiate Imputer.

        Parameters
        ----------
        model : Model
            Model.
        data : pd.DataFrame
            Data with missing data labeled as np.nan.
        factors: bool
            If True, factors are estimated. The default is True.

        Returns
        -------
        None.

        """
        self.mod = model
        self.mx_data_imp = data.copy()
        self.n_param_missing = 0
        self.factors = factors
        desc = model.description
        self.dict_effects[self.symb_missing] = self.effect_missing
        super().__init__(desc, mimic_lavaan=model.mimic_lavaan)
        self.objectives = {'ML': (self.obj_ml, self.grad_ml)}
        self.load_starting_values()

    def finalize_variable_classification(self):
        """
        Finalize variable classification.

        Reorders variables for better visual fancyness and does extra
        model-specific variable respecification.
        Returns
        -------
        None.

        """
        super().finalize_variable_classification()
        if self.factors:
            self.vars['observed'] += sorted(self.vars['latent'])

    def setup_matrices(self):
        """
        Initialize base matrix structures of the model.

        Returns
        -------
        None.

        """
        super().setup_matrices()
        for i, f in enumerate(self.start_rules):
            name = f.__name__
            if not name.endswith('_imp'):
                self.start_rules[i] = getattr(startingvalues, name + '_imp')

    def build_data_imp(self):
        """
        Build model-implied model imputed data matrix.

        Returns
        -------
        mx : np.ndarrau
            Model-implied data matrix.
        names : tuple
            Row and column names.

        """
        mx = self.mx_data_imp[self.vars['observed']].copy()
        names = (list(mx.index), list(mx.columns))
        return mx.values, names

    def preprocess_effects(self, effects: dict):
        """
        Run a routine just before effects are applied.

        Used to apply covariances to model.
        Parameters
        -------
        effects : dict
            Mapping opcode->lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        super().preprocess_effects(effects)
        for _, lvals in effects.items():
            for lval, rvals in lvals.items():
                for rval in rvals:
                    rvals[rval] = self.symb_starting_values
        missing = effects[self.symb_missing]
        obs = self.vars['observed']
        mx = self.mx_data_imp
        for v in obs:
            if v not in mx.columns:
                mx[v] = np.nan
        mx = mx[obs]
        inds = list(mx.index)
        for i, j in zip(*np.where(np.isnan(mx))):
            v = obs[j]
            missing[inds[i]][v] = None

    def effect_missing(self, items: dict):
        """
        Work through missing operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        mx, (rows, cols) = self.mx_data_imp, self.names_data_imp
        for lv, rvs in items.items():
            i = rows.index(lv)
            for rv in rvs:
                j = cols.index(rv)
                ind = (i, j)
                self.n_param_missing += 1
                name = f'_m{self.n_param_missing}'
                self.add_param(name, matrix=mx, indices=ind, start=None,
                               active=True, symmetric=False,
                               bound=(None, None))

    def fit(self, solver='SLSQP', clean_slate=False):
        """
        Perform data imputation.

        Parameters
        ----------
        solver: str
            Solver to use. The default is 'SLSQP'.

        Returns
        ----------
        SolverResult:
            Result of optimizaiton.
        """
        if clean_slate or not hasattr(self, 'param_vals'):
            self.prepare_params()
        obj, grad = self.get_objective('ML')
        solver = Solver(solver, obj, grad, self.param_vals)
        res = solver.solve()
        self.update_matrices(res.x)
        return res

    def prepare_params(self):
        """
        Prepare structures for effective optimization routines.

        Returns
        -------
        None.

        """
        super().prepare_params()
        self.mx_sigma = self.calc_sigma()[0]
        try:
            self.mx_sigma_inv = chol_inv(self.mx_sigma)
        except np.linalg.LinAlgError:
            self.mx_sigma_inv = np.linalg.pinv(self.mx_sigma)

    def calc_data_grad(self):
        """
        Calculate model-implied data gradient.

        Returns
        -------
        list
            List of gradient values.

        """
        return self.mx_diffs

    def obj_ml(self, x: np.ndarray):
        """
        Calculate ML objective value.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Loglikelihood value.

        """
        self.update_matrices(x)
        data = self.mx_data_imp
        return np.einsum('ij,jk,ik->', data, self.mx_sigma_inv, data)

    def grad_ml(self, x: np.ndarray):
        """
        Gradient of ML objective function.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of GLS.

        """
        self.update_matrices(x)
        data_grad = self.calc_data_grad()
        t = self.mx_sigma_inv @ self.mx_data_imp.T
        return 2 * np.array([np.einsum('ij,ji->', g[4], t)
                             for g in data_grad])

    def get_fancy(self):
        """
        Returns imputed data in DataFrame form.

        Returns
        -------
        pd.DataFrame
            DataFrame with imputed data and factor scores.

        """
        data = pd.DataFrame(self.mx_data_imp, index=self.names_data_imp[0],
                            columns=self.names_data_imp[1])
        return data

    def operation_start(self, operation):
        pass

    def operation_bound(self, operation):
        pass

    def operation_constraint(self, operation):
        pass


class ImputerMeans(ModelMeans):
    """ModelMeans for missing data imputation."""

    symb_missing = '@'

    matrices_names = tuple(list(ModelMeans.matrices_names) +\
                           ['data_imp', 'g_imp'])

    def __init__(self, model: ModelMeans, data: pd.DataFrame, factors=True):
        """
        Instantiate ImputerMeans.

        Parameters
        ----------
        model : ModelMeans
            Model with meanstructure.
        data : pd.DataFrame
            Data with missing data labeled as np.nan.
        factors: bool
            If True, factors are estimated. The default is True.

        Returns
        -------
        None.

        """
        self.mod = model
        self.mx_data_imp = data.copy()
        if model.intercepts:
            data = data.copy()
            data['1'] = 1.0
        t = [v for v in model.vars['observed_exogenous']
             if v in data.columns]
        self.mx_g = data[t].copy()
        self.n_param_missing = 0
        self.factors = factors
        desc = model.description
        self.dict_effects[self.symb_missing] = self.effect_missing
        super().__init__(desc, mimic_lavaan=model.mimic_lavaan,
                         intercepts=model.intercepts)
        self.objectives = {'ML': (self.obj_ml, self.grad_ml)}
        self.load_starting_values()

    def finalize_variable_classification(self):
        """
        Finalize variable classification.

        Reorders variables for better visual fancyness and does extra
        model-specific variable respecification.
        Returns
        -------
        None.

        """
        super().finalize_variable_classification()
        if self.factors:
            self.vars['observed'] += sorted(self.vars['latent'])

    def setup_matrices(self):
        """
        Initialize base matrix structures of the model.

        Returns
        -------
        None.

        """
        super().setup_matrices()
        for i, f in enumerate(self.start_rules):
            name = f.__name__
            if not name.endswith('_imp'):
                self.start_rules[i] = getattr(startingvalues, name + '_imp')

    def build_data_imp(self):
        """
        Build model-implied model imputed data matrix.

        Returns
        -------
        mx : np.ndarray
            Model-implied data matrix.
        names : tuple
            Row and column names.

        """
        mx = self.mx_data_imp[self.vars['observed']].copy()
        names = (list(mx.columns), list(mx.index))
        return mx.values.T, names

    def build_g_imp(self):
        """
        Build model-implied model imputed data G matrix.

        Returns
        -------
        mx : np.ndarray
            Model-implied G matrix.
        names : tuple
            Row and column names.

        """
        mx = self.mx_g[self.vars['observed_exogenous']].copy()
        names = (list(mx.columns), list(mx.index))
        return mx.values.T, names

    def preprocess_effects(self, effects: dict):
        """
        Run a routine just before effects are applied.

        Used to apply covariances to model.
        Parameters
        -------
        effects : dict
            Mapping opcode->lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        super().preprocess_effects(effects)
        for _, lvals in effects.items():
            for lval, rvals in lvals.items():
                for rval in rvals:
                    rvals[rval] = self.symb_starting_values
        missing = effects[self.symb_missing]
        obs = self.vars['observed']
        mx = self.mx_data_imp
        for v in obs:
            if v not in mx.columns:
                mx[v] = np.nan
        mx = mx[obs]
        inds = list(mx.index)
        for i, j in zip(*np.where(np.isnan(mx))):
            v = obs[j]
            missing[inds[i]][v] = None

        obs = self.vars['observed_exogenous']
        mx = self.mx_g
        for v in obs:
            if v not in mx.columns:
                mx[v] = np.nan
        mx = mx[obs]
        inds = list(mx.index)
        for i, j in zip(*np.where(np.isnan(mx))):
            v = obs[j]
            missing[inds[i]][v] = None

    def effect_missing(self, items: dict):
        """
        Work through missing operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        obs_exo1 = self.vars['observed_exogenous_1']
        obs_exo2 = self.vars['observed_exogenous_2']
        for lv, rvs in items.items():
            for rv in rvs:
                if lv in obs_exo1:
                    mx = self.mx_g1
                    rows, cols = self.names_g1_imp
                elif rv in obs_exo2:
                    mx = self.mx_g2
                    rows, cols = self.names_g2_imp
                else:
                    mx, (rows, cols) = self.mx_data_imp, self.names_data_imp
                i = rows.index(rv)
                j = cols.index(lv)
                ind = (i, j)
                self.n_param_missing += 1
                name = f'_m{self.n_param_missing}'
                self.add_param(name, matrix=mx, indices=ind, start=None,
                               active=True, symmetric=False,
                               bound=(None, None))

    def fit(self, solver='SLSQP', clean_slate=False):
        """
        Perform data imputation.

        Parameters
        ----------
        solver: str
            Solver to use. The default is 'SLSQP'.

        Returns
        ----------
        SolverResult:
            Result of optimizaiton.
        """
        if not clean_slate or not hasattr(self, 'param_vals'):
            self.prepare_params()
        obj, grad = self.get_objective('ML')
        solver = Solver(solver, obj, grad, self.param_vals)
        res = solver.solve()
        self.update_matrices(res.x)
        return res

    def prepare_params(self):
        """
        Prepare structures for effective optimization routines.

        Returns
        -------
        None.

        """
        super().prepare_params()
        self.mx_sigma = self.calc_sigma()[0]
        try:
            self.mx_sigma_inv = chol_inv(self.mx_sigma)
        except np.linalg.LinAlgError:
            self.mx_sigma_inv = np.linalg.pinv(self.mx_sigma)
        i = np.identity(self.mx_beta.shape[0])
        self.mx_m = self.mx_lambda @ np.linalg.inv(i - self.mx_beta)
        self.mx_g = self.mx_g_imp
        self.mx_mg = self.mx_m @ self.mx_gamma1 + self.mx_gamma2

    def calc_data_grad(self):
        """
        Calculate model-implied data gradient.

        Returns
        -------
        list
            List of gradient values.

        """
        grad = list()
        for mxs in self.mx_diffs:
            g = np.float32(0.0)
            if mxs[-2] is not None:  # data_imp
                g += mxs[-2]
            if mxs[-1] is not None:  # g1
                g -= self.mx_mg @ mxs[-1]
            grad.append(g)
        return grad

    def obj_ml(self, x: np.ndarray):
        """
        Calculate ML objective value.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Loglikelihood value.

        """
        self.update_matrices(x)
        center = self.mx_data_imp - self.calc_mean(self.mx_m)
        return np.einsum('ji,jk,ki->', center, self.mx_sigma_inv, center)

    def grad_ml(self, x: np.ndarray):
        """
        Gradient of ML objective function.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of ML.

        """
        self.update_matrices(x)
        data_grad = self.calc_data_grad()
        center = self.mx_data_imp - self.calc_mean(self.mx_m)
        t = self.mx_sigma_inv @ center
        return 2 * np.array([np.einsum('ji,ji->', g, t)
                             for g in data_grad])

    def get_fancy(self):
        """
        Returns imputed data in DataFrame form.

        Returns
        -------
        pd.DataFrame
            DataFrame with imputed data and factor scores.

        """
        data = pd.DataFrame(self.mx_data_imp, index=self.names_data_imp[0],
                            columns=self.names_data_imp[1])
        g = pd.DataFrame(self.mx_g_imp, index=self.names_g_imp[0],
                            columns=self.names_g_imp[1])
        if '1' in g.index:
            g.drop('1', inplace=True)
        res = pd.concat([data, g]).T
        var = self.mod.vars
        t = sorted(var['all'] - var['latent']) + sorted(var['latent'])
        return res[t]

    def operation_start(self, operation):
        pass

    def operation_bound(self, operation):
        pass

    def operation_constraint(self, operation):
        pass

class ImputerEffects(ModelEffects):
    """ModelEffects for missing data imputation."""

    symb_missing = '@'

    matrices_names = tuple(list(ModelEffects.matrices_names) +\
                           ['data_imp', 'g_imp'])

    def __init__(self, model: ModelMeans, data: pd.DataFrame, group: str, 
                 k=None, factors=True):
        """
        Instantiate ImputerMeans.

        Parameters
        ----------
        model : ModelMeans
            Model with meanstructure.
        data : pd.DataFrame
            Data with missing data labeled as np.nan.
        k : pd.DataFrame
            Kinship between individuals matrix. If None, identity is assumed.
            The default is None.
        factors: bool
            If True, factors are estimated. The default is True.

        Returns
        -------
        None.

        """
        self.mod = model
        if model.intercepts:
            data = data.copy()
            data['1'] = 1.0
        self.mx_data_orig = data.copy()
        self.mx_data_imp = data
        t = [v for v in model.vars['observed_exogenous']
             if v in data.columns]
        self.mx_g = data[t].copy()
        self.n_param_missing = 0
        self.factors = factors
        desc = model.description
        self.dict_effects[self.symb_missing] = self.effect_missing
        super().__init__(desc, mimic_lavaan=model.mimic_lavaan,
                         intercepts=model.intercepts)
        self.objectives = {'ML': (self.obj_ml, None)}
        self.load_starting_values()
        self.load(group, k=k)

    def load(self, group: str, k=None):
        """
        Load kinship K matrix.

        Parameters
        ----------
        group : str
            Name of column with group labels.
        k : pd.DataFrame
            Covariance matrix across rows, i.e. kinship matrix. If None,
            identity is assumed. The default is None.

        KeyError
            Rises when there are missing variables from the data.
        Exception
            Rises when group parameter is None.
        Returns
        -------
        None.

        """
        obs = self.vars['observed']
        data = self.mx_data_orig
        grs = data[group]
        p_names = list(grs.unique())
        p, n = len(p_names), data.shape[0]
        if k is None:
            k = np.identity(p)
        elif k.shape[0] != p:
            raise Exception("Dimensions of K don't match number of groups.")
        z = np.zeros((n, p))
        for i, germ in enumerate(grs):
            j = p_names.index(germ)
            z[i, j] = 1.0
        if type(k) is pd.DataFrame:
            try:
                k = k.loc[p_names, p_names].values
            except KeyError:
                raise KeyError("Certain groups in K differ from those "\
                                "provided in a dataset.")
        self.mx_g = data[self.vars['observed_exogenous']].values.T
        if len(self.mx_g.shape) != 2:
            self.mx_g = self.mx_g[np.newaxis, :]
        g = self.mx_g
        self.num_m = self.vars['observed']
        zkz = z @ k @ z.T
        c = np.linalg.inv(np.identity(self.mx_beta.shape[0]) - self.mx_beta)
        m = self.mx_lambda @ c
        sigma = m @ self.mx_psi @ m.T + self.mx_theta
        self.mx_m = m
        m = len(obs)
        r = n * (np.ones((m, m)) * self.mx_v[0, 0] + sigma)\
            + np.trace(zkz) * self.mx_d
        w = np.ones((n, n)) * np.trace(sigma) + zkz * np.trace(self.mx_d) +\
            np.identity(n) * self.mx_v[0, 0] * m
        self.mx_r_inv = chol_inv(r)
        self.mx_w_inv = chol_inv(w)

    def finalize_variable_classification(self):
        """
        Finalize variable classification.

        Reorders variables for better visual fancyness and does extra
        model-specific variable respecification.
        Returns
        -------
        None.

        """
        super().finalize_variable_classification()
        if self.factors:
            self.vars['observed'] += sorted(self.vars['latent'])

    def setup_matrices(self):
        """
        Initialize base matrix structures of the model.

        Returns
        -------
        None.

        """
        super().setup_matrices()
        for i, f in enumerate(self.start_rules):
            name = f.__name__
            if not name.endswith('_imp'):
                self.start_rules[i] = getattr(startingvalues, name + '_imp')

    def build_data_imp(self):
        """
        Build model-implied model imputed data matrix.

        Returns
        -------
        mx : np.ndarray
            Model-implied data matrix.
        names : tuple
            Row and column names.

        """
        mx = self.mx_data_imp[self.vars['observed']].copy()
        names = (list(mx.columns), list(mx.index))
        return mx.values.T, names

    def build_g_imp(self):
        """
        Build model-implied model imputed data G matrix.

        Returns
        -------
        mx : np.ndarray
            Model-implied G matrix.
        names : tuple
            Row and column names.

        """
        mx = self.mx_g[self.vars['observed_exogenous']].copy()
        names = (list(mx.columns), list(mx.index))
        return mx.values.T, names

    def preprocess_effects(self, effects: dict):
        """
        Run a routine just before effects are applied.

        Used to apply covariances to model.
        Parameters
        -------
        effects : dict
            Mapping opcode->lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        super().preprocess_effects(effects)
        for _, lvals in effects.items():
            for lval, rvals in lvals.items():
                for rval in rvals:
                    rvals[rval] = self.symb_starting_values
        missing = effects[self.symb_missing]
        obs = self.vars['observed']
        mx = self.mx_data_imp
        for v in obs:
            if v not in mx.columns:
                mx[v] = np.nan
        mx = mx[obs]
        inds = list(mx.index)
        for i, j in zip(*np.where(np.isnan(mx))):
            v = obs[j]
            missing[inds[i]][v] = None

        obs = self.vars['observed_exogenous']
        mx = self.mx_g
        for v in obs:
            if v not in mx.columns:
                mx[v] = np.nan
        mx = mx[obs]
        inds = list(mx.index)
        for i, j in zip(*np.where(np.isnan(mx))):
            v = obs[j]
            missing[inds[i]][v] = None

    def effect_missing(self, items: dict):
        """
        Work through missing operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        obs_exo1 = self.vars['observed_exogenous_1']
        obs_exo2 = self.vars['observed_exogenous_2']
        for lv, rvs in items.items():
            for rv in rvs:
                if lv in obs_exo1:
                    mx = self.mx_g1
                    rows, cols = self.names_g1_imp
                elif rv in obs_exo2:
                    mx = self.mx_g2
                    rows, cols = self.names_g2_imp
                else:
                    mx, (rows, cols) = self.mx_data_imp, self.names_data_imp
                i = rows.index(rv)
                j = cols.index(lv)
                ind = (i, j)
                self.n_param_missing += 1
                name = f'_m{self.n_param_missing}'
                self.add_param(name, matrix=mx, indices=ind, start=None,
                               active=True, symmetric=False,
                               bound=(None, None))

    def fit(self, solver='SLSQP', clean_slate=False):
        """
        Perform data imputation.

        Parameters
        ----------
        solver: str
            Solver to use. The default is 'SLSQP'.

        Returns
        ----------
        SolverResult:
            Result of optimizaiton.
        """
        if clean_slate or not hasattr(self, 'param_vals'):
            self.prepare_params()
        obj, grad = self.get_objective('ML')
        solver = Solver(solver, obj, grad, self.param_vals)
        res = solver.solve()
        self.update_matrices(res.x)
        return res

    def prepare_params(self):
        """
        Prepare structures for effective optimization routines.

        Returns
        -------
        None.

        """
        super().prepare_params()

    def calc_data_grad(self):
        """
        Calculate model-implied data gradient.

        Returns
        -------
        list
            List of gradient values.

        """
        grad = list()
        for mxs in self.mx_diffs:
            g = np.float32(0.0)
            if mxs[-2] is not None:  # data_imp
                g += mxs[-2]
            if mxs[-1] is not None:  # g1
                g -= self.mx_mg @ mxs[-1]
            grad.append(g)
        return grad

    def obj_ml(self, x: np.ndarray):
        """
        Calculate ML objective value.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Loglikelihood value.

        """
        self.update_matrices(x)
        center = self.mx_data_imp - self.calc_mean(self.mx_m)
        r, w = self.mx_r_inv, self.mx_w_inv
        return np.einsum('ij,jk->', w @ center.T, r @ center)

    def grad_ml(self, x: np.ndarray):
        """
        Gradient of ML objective function.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of ML.

        """
        self.update_matrices(x)
        data_grad = self.calc_data_grad()
        center = self.mx_data_imp - self.calc_mean(self.mx_m)
        t = self.mx_sigma_inv @ center
        return 2 * np.array([np.einsum('ji,ji->', g, t)
                             for g in data_grad])

    def get_fancy(self):
        """
        Returns imputed data in DataFrame form.

        Returns
        -------
        pd.DataFrame
            DataFrame with imputed data and factor scores.

        """
        data = pd.DataFrame(self.mx_data_imp, index=self.names_data_imp[0],
                            columns=self.names_data_imp[1])
        g = pd.DataFrame(self.mx_g_imp, index=self.names_g_imp[0],
                            columns=self.names_g_imp[1])
        if '1' in g.index:
            g.drop('1', inplace=True)
        res = pd.concat([data, g]).T
        var = self.mod.vars
        t = sorted(var['all'] - var['latent']) + sorted(var['latent'])
        return res[t]

    def operation_start(self, operation):
        pass

    def operation_bound(self, operation):
        pass

    def operation_constraint(self, operation):
        pass


def get_imputer(self):
    """
    Retrieve an appropriate Imputer instance.

    Parameters
    ----------
    self : Model or ModelMeans
        Model.

    Returns
    -------
    Imputer, ImputerMeans or ImputerEffects.

    """
    if type(self) is Model:
        return Imputer
    elif type(self) is ModelMeans:
        return ImputerMeans
    elif type(self) is ModelEffects:
        return ImputerEffects



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examples for semopy."""
from . import political_democracy
from . import univariate_regression
from . import univariate_regression_many
from . import multivariate_regression
from . import holzinger39
from . import example_rf
from . import example_article


# -*- coding: utf-8 -*-
"""Generalized Random Effects SEM."""
from .utils import chol_inv, chol_inv2, cov
from scipy.linalg import solve_sylvester
from .model_means import ModelMeans
from itertools import combinations
from functools import partial
from . import startingvalues
from copy import deepcopy
import pandas as pd
import numpy as np


class ModelGeneralizedEffects(ModelMeans):
    """
    Generalized Random Effects model.
    
    Generalized Random Effects SEM is a generalization of ModelEffects in a
    sense, that it allows for an arbitrary number of random effects, and also
    it allows to introduce parametic covariance-between-observations marices.
    The latter can be thought of as in context of time-series or spatial data.
    """

    def __init__(self, description: str, effects: tuple, mimic_lavaan=False,
                 baseline=False, cov_diag=False, intercepts=True):
        """
        Instantiate Generalized Random Effects SEM model.

        Parameters
        ----------
        description : str
            Model description in semopy syntax.
        effects : tuple, EffectBase
            A list of Effects or a single effect.
        mimic_lavaan: bool
            If True, output variables are correlated and not conceptually
            identical to indicators. lavaan treats them that way, but it's
            less computationally effective. The default is False.
        baseline : bool
            If True, the model will be set to baseline model.
            Baseline model here is an independence model where all variables
            are considered to be independent with zero covariance. Only
            variances are estimated. The default is False.
        cov_diag : bool
            If cov_diag is True, then there are no covariances parametrised
            unless explicitly specified. The default is False.
        intercepts: bool
            If True, intercepts are also modeled. Intercept terms can be
            accessed via "1" symbol in a regression equation, i.e. x1 ~ 1. The
            default is False.

        Returns
        -------
        None.

        """
        if type(effects) not in (list, tuple):
            effects = (effects, )
        self.effects = effects
        self.symbs_rf = [f'~{i+1}~' for i in range(len(effects))]
        matrices = list(self.matrices_names)
        for i, symb in enumerate(self.symbs_rf):
            name = f'd{i+1}'
            setattr(self, f'build_{name}', self.build_d)
            setattr(self, f'start_{name}', startingvalues.start_d)
            matrices.append(name)
            f = partial(self.effect_rf_covariance, mx=name)
            self.dict_effects[symb] = f
        self.matrices_names = tuple(matrices)
        super().__init__(description, mimic_lavaan=False, baseline=baseline,
                         cov_diag=cov_diag, intercepts=intercepts)
        self.objectives = {'FIML': (self.obj_matnorm, self.grad_matnorm)}
    
    def preprocess_effects(self, effects: dict):
        """
        Run a routine just before effects are applied.

        Used to apply random effect variance
        Parameters
        -------
        effects : dict
            Mapping opcode->lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        super().preprocess_effects(effects)
        obs = self.vars['observed']
        for i, effect in enumerate(self.effects):
            symb = self.symbs_rf[i]
            mode = effect.d_mode
            if mode in ('diag', 'full'):
                for v in obs:
                    t = effects[symb][v]
                    if v not in t:
                        t[v] = None
                if mode == 'full':
                    for a, b in combinations(obs, 2):
                        t = effects[symb][a]
                        tt = effects[symb][b]
                        if (v not in t) and (v not in tt):
                            t[b] = None
            else:
                if mode != 'scale':
                    raise Exception(f'Unknown mode "{mode}".')
                param = f'paramD{i + 1}'
                for v in obs:
                    t = effects[symb][v][v] = param

    def load(self, data, cov=None, clean_slate=False, n_samples=None,
             **kwargs):
        """
        Load dataset.

        Parameters
        ----------
        data : pd.DataFrame
            Data with columns as variables.
        cov : pd.DataFrame, optional
            Pre-computed covariance/correlation matrix. Used only for variance
            starting values. The default is None.
        clean_slate : bool, optional
            If True, resets parameters vector. The default is False.
        n_samples : int, optional
            Redunant for ModelEffects. The default is None.
        **kwargs : dict
            Extra arguments are sent to Effects.

        KeyError
            Rises when there are missing variables from the data.
        Exception
            Rises when group parameter is None.
        Returns
        -------
        None.

        """
        if data is None:
            if not hasattr(self, 'mx_data'):
                raise Exception("Data must be provided.")
            if clean_slate:
                self.prepare_params()
            return
        else:
            data = data.copy()
        obs = self.vars['observed']
        exo = self.vars['observed_exogenous']
        if self.intercepts:
            data['1'] = 1.0
        cols = data.columns
        missing = (set(obs) | set(exo)) - set(set(cols))
        if missing:
            t = ', '.join(missing)
            raise KeyError(f'Variables {t} are missing from data.')
        self.load_data(data, covariance=cov, **kwargs)
        self.load_starting_values()
        if clean_slate or not hasattr(self, 'param_vals'):
            self.prepare_params()


    def prepare_params(self):
        """
        Prepare structures for effective optimization routines.

        Returns
        -------
        None.

        """
        super().prepare_params()
        extra = np.array([])
        ranges = list()
        a = len(self.param_vals)
        for effect in self.effects:
            extra = np.append(extra, effect.parameters)
            b = a + len(effect.parameters)
            ranges.append((a, b))
            a = b
        self.param_vals = np.append(self.param_vals, extra)
        self.effects_param_ranges = ranges

    def update_matrices(self, params: np.ndarray):
        """
        Update all matrices from a parameter vector.

        Parameters
        ----------
        params : np.ndarray
            Vector of parameters.

        Returns
        -------
        None.

        """
        super().update_matrices(params)
        for effect, (a, b) in zip(self.effects, self.effects_param_ranges):
            effect.parameters[:] = params[a:b]

    def build_d(self):
        """
        D matrix is a covariance matrix for random effects across columns.

        Returns
        -------
        np.ndarray
            Matrix.
        tuple
            Tuple of rownames and colnames.

        """
        names = self.vars['observed']
        n = len(names)
        mx = np.zeros((n, n))
        return mx, (names, names)

    def effect_rf_covariance(self, items: dict, mx: str):
        """
        Work through random effects covariance operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.
        mx : str
            Name of the D matrix.

        Returns
        -------
        None.

        """
        rows, cols = getattr(self, f'names_{mx}')
        mx = getattr(self, f'mx_{mx}')
        for lv, rvs in items.items():
            for rv, mult in rvs.items():
                name = None
                try:
                    val = float(mult)
                    active = False
                except (TypeError, ValueError):
                    active = True
                    if mult is not None:
                        if mult != self.symb_starting_values:
                            name = mult
                        else:
                            active = False
                    val = None
                if name is None:
                    self.n_param_cov += 1
                    name = '_c%s' % self.n_param_cov
                i, j = rows.index(lv), cols.index(rv)
                ind = (i, j)
                if i == j:
                    bound = (0, None)
                    symm = False
                else:
                    if self.baseline:
                        continue
                    bound = (None, None)
                    symm = True
                self.add_param(name, matrix=mx, indices=ind, start=val,
                               active=active, symmetric=symm, bound=bound)

    def get_bounds(self):
        """
        Get bound constraints on parameters.

        Returns
        -------
        list
            List of tuples specifying bounds.

        """
        b = super().get_bounds()
        for effect in self.effects:
            b.extend(effect.get_bounds())
        return b


    def fit(self, data=None, cov=None, obj='ML', solver='SLSQP', groups=None,
            clean_slate=False, **kwargs):
        """
        Fit model to data.

        Parameters
        ----------
        data : pd.DataFrame, optional
            Data with columns as variables. The default is None.
        cov : pd.DataFrame, optional
            Pre-computed covariance/correlation matrix. The default is None.
        obj : str, optional
            Objective function to minimize. Possible values are 'REML', "ML".
            The default is 'ML'.
        solver : TYPE, optional
            Optimizaiton method. Currently scipy-only methods are available.
            The default is 'SLSQP'.
        groups : list, optional
            Groups of size > 1 to center across. The default is None.
        clean_slate : bool, optional
            If False, successive fits will be performed with previous results
            as starting values. If True, parameter vector is reset each time
            prior to optimization. The default is False.

        Raises
        ------
        NotImplementedError
            Rises when unknown objective name is passed.

        Returns
        -------
        SolverResult
            Information on optimization process.

        """
        if obj == 'ML':
            res = super().fit(data=data, cov=cov, obj='ML', solver=solver,
                              groups=groups, clean_slate=clean_slate, **kwargs)
            return res
        else:
            raise NotImplementedError(f"Unknown method {obj}.")

    '''
    ---------------------Preparing structures for a more-----------------------
    ------------------------efficient computations-----------------------------
    '''

    def load_data(self, data: pd.DataFrame, covariance=None, **kwargs):
        """
        Load dataset from data matrix.

        Parameters
        ----------
        data : pd.DataFrame
            Dataset with columns as variables and rows as observations.
        covariance : pd.DataFrame, optional 
            Custom covariance matrix. The default is None.
         **kwargs : dict
            Extra arguments are sent to Effects.

        Returns
        -------
        None.

        """
        obs = self.vars['observed']
        self.mx_g = data[self.vars['observed_exogenous']].values.T
        if len(self.mx_g.shape) != 2:
            self.mx_g = self.mx_g[np.newaxis, :]
        self.mx_data = data[obs].values.T
        self.n_obs, self.n_samples = self.mx_data.shape
        self.num_m = len(set(self.vars['observed']) - self.vars['latent'])
        self.load_cov(covariance[obs].loc[obs]
                      if covariance is not None else cov(self.mx_data.T))
        d_matrices = list()
        for i, effect in enumerate(self.effects):
            effect.load(i, self, data, **kwargs)
            d = getattr(self, f'mx_d{i + 1}')
            d_matrices.append(d)
        self.mxs_d = d_matrices
        self.mx_identity = np.identity(self.n_samples)

    '''
    ----------------------------LINEAR ALGEBRA PART---------------------------
    ----------------------The code below is responsible-----------------------
    ------------------for covariance structure computations-------------------
    '''

    def calc_l(self, sigma=None, k=None):
        """
        Calculate covariance across columns matrix T.
        
        Parameters
        ----------
        sigma: np.ndarray, optional
            Sigma covariance matrix as returned by calc_sigma. Although there
            is no meaningful concept of Sigma matrix in ModelEffects, it is
            still computationally convenient to separate it into an extra
            element. If None, then it will computed automatically. The default
            is None.
        k: tuple, optional
            List of K matrices as returned by calc_k by Effects. If None, then
            calculated in place. The default is None.
        Returns
        -------
        np.ndarray
            Covariance across columns (variables) matrix T.

        """
        if sigma is None:
            sigma, _ = self.calc_sigma()
        if k is None:
            k = self.calc_ks()
        n = self.n_samples
        return sum(np.trace(k) * d for d, k in zip(self.mxs_d, k)) + n * sigma

    def calc_l_grad(self, sigma=None, sigma_grad=None, k=None, k_grad=None):
        """
        Calculate gradient of covariance across columns matrix T.
        
        Parameters
        ----------
        sigma: np.ndarray, optional
            Sigma covariance matrix as returned by calc_sigma. Although there
            is no meaningful concept of Sigma matrix in ModelEffects, it is
            still computationally convenient to separate it into an extra
            element. If None, then it will computed automatically. The default
            is None.
        sigma_grad: List[np.ndarray], optional
            List of Sigma derivatives as returned by calc_sigma_grad. If None,
            then will be computed in place. The default is None.
        k: tuple, optional
            List of K matrices as returned by calc_k by Effects. If None, then
            calculated in place. The default is None.
        sigma_grad: List[List[np.ndarray]], optional
            List of K gradients as returned by calc_k_grad of Effect. If None,
            then will be computed in place. The default is None.
        Returns
        -------
        List[np.ndarray]
            List of derivatives ofcovariance across columns (variables) matrix
            wrt to model parameters.

        """
        if sigma is None:
            sigma, (m, c) = self.calc_sigma()
            if sigma_grad is None:
                sigma_grad = self.calc_sigma_grad(m, c)
        if k is None:
            k = self.calc_ks()
            if k_grad is None:
                k_grad= self.calc_ks_grad()
        k = list(map(np.trace, k))
        k_grad = list(map(np.trace, k_grad))
        grad = list()
        n = self.n_samples
        for g, df in zip(sigma_grad, self.mx_diffs):
            g = g * n
            for i in range(6, len(df)):
                if df[i] is not None:
                    g += df[i] * k[i - 6]
            grad.append(g)
        c = 0
        for i, (a, b) in enumerate(self.effects_param_ranges):
            d = self.mxs_d[i]
            for _ in range(b - a):
                grad.append(d * k_grad[c])
                c += 1
        return grad

    def calc_t(self, sigma=None, k=None):
        """
        Calculate covariance across rows matrix L.
        
        Parameters
        ----------
        sigma: np.ndarray, optional
            Sigma covariance matrix as returned by calc_sigma. Although there
            is no meaningful concept of Sigma matrix in ModelEffects, it is
            still computationally convenient to separate it into an extra
            element. If None, then it will computed automatically. The default
            is None.
        k: tuple, optional
            List of K matrices as returned by calc_k by Effects. If None, then
            calculated in place. The default is None.
        Returns
        -------
        np.ndarray
            Covariance across rows (observations) matrix L.

        """
        if sigma is None:
            sigma, _ = self.calc_sigma()
        if k is None:
            k = self.calc_ks()
        s = self.mx_identity * np.trace(sigma)
        return sum(np.trace(d) * k for d, k in zip(self.mxs_d, k)) + s

    def calc_t_grad(self, sigma=None, sigma_grad=None, k=None, k_grad=None):
        """
        Calculate gradient of covariance across rows matrix L.
        
        Parameters
        ----------
        sigma: np.ndarray, optional
            Sigma covariance matrix as returned by calc_sigma. Although there
            is no meaningful concept of Sigma matrix in ModelEffects, it is
            still computationally convenient to separate it into an extra
            element. If None, then it will computed automatically. The default
            is None.
        sigma_grad: List[np.ndarray], optional
            List of Sigma derivatives as returned by calc_sigma_grad. If None,
            then will be computed in place. The default is None.
        k: tuple, optional
            List of K matrices as returned by calc_k by Effects. If None, then
            calculated in place. The default is None.
        sigma_grad: List[List[np.ndarray]], optional
            List of K gradients as returned by calc_k_grad of Effect. If None,
            then will be computed in place. The default is None.
        Returns
        -------
        List[np.ndarray]
            List of derivatives ofcovariance across rows (observations) matrix
            wrt to model parameters.

        """
        if sigma is None:
            sigma, (m, c) = self.calc_sigma()
            if sigma_grad is None:
                sigma_grad = self.calc_sigma_grad(m, c)
        if k is None:
            k = self.calc_ks()
            if k_grad is None:
                k_grad= self.calc_ks_grad()
        grad = list()
        for g, df in zip(sigma_grad, self.mx_diffs):
            try:
                g = np.trace(g) * self.mx_identity
            except ValueError:
                g = 0.0
            for i in range(6, len(df)):
                if df[i] is not None:
                    g += np.trace(df[i]) * k[i - 6]
            grad.append(g)
        c = 0
        for i, (a, b) in enumerate(self.effects_param_ranges):
            d = np.trace(self.mxs_d[i])
            for _ in range(b - a):
                grad.append(d * k_grad[c])
                c += 1
        return grad

    def calc_ks(self):
        return [effect.calc_k(self) for effect in self.effects]

    def calc_ks_grad(self):
        grad = list()
        for effect in self.effects:
            grad.extend(effect.calc_k_grad(self))
        return grad
    
    def calc_mean_grad(self, m=None, c=None):
        if m is None:
            m, c = self.calc_sigma()[1]
        grad = super().calc_mean_grad(m, c)
        n = len(grad)
        p = len(self.param_vals)
        grad.extend([np.float(0.0)] * (p - n))
        return grad
    
    '''
    ------------------Matrix Variate Normal Maximum Likelihood-----------------
    '''
    def obj_matnorm(self, x: np.ndarray):
        """
        Loglikelihood of matrix-variate normal distribution.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Loglikelihood (constants omitted).

        """
        self.update_matrices(x)
        sigma, (m, _) = self.calc_sigma()
        k = self.calc_ks()
        l = self.calc_l(sigma, k)
        t = self.calc_t(sigma, k)
        try:
            l_inv, l_logdet = chol_inv2(l)
            t_inv, t_logdet = chol_inv2(t)
        except np.linalg.LinAlgError:
            return np.nan
        center = self.mx_data - self.calc_mean(m)
        tr_l = np.trace(l)
        a = tr_l * np.einsum('ij,ji', l_inv @ center, t_inv @ center.T)
        m = self.num_m
        n = self.n_samples
        return a + n * l_logdet + m * t_logdet - n * m * np.log(tr_l)
        
    def grad_matnorm(self, x: np.ndarray):
        grad = np.zeros_like(x)
        self.update_matrices(x)
        sigma, (m, c) = self.calc_sigma()
        k = self.calc_ks()
        try:
            l = self.calc_l(sigma, k)
            l_inv = chol_inv(l)
            t = self.calc_t(sigma, k)
            t_inv = chol_inv(t)
        except np.linalg.LinAlgError:
            grad[:] = np.inf
            return grad
        mean_grad = self.calc_mean_grad(m, c)
        sigma_grad = self.calc_sigma_grad(m, c)
        k_grad = self.calc_ks_grad()
        l_grad = self.calc_l_grad(sigma, sigma_grad, k, k_grad)
        t_grad = self.calc_t_grad(sigma, sigma_grad, k, k_grad)
        center = self.mx_data - self.calc_mean(m)
        m = self.num_m
        n = self.n_samples
        c0 = t_inv @ center.T @ l_inv
        # a1 = l_inv @ center @ t_inv
        c1 = c0 @ center
        c2 = center @ c0
        big_tr = np.trace(c1)
        tr_l = np.trace(l)
        for i, (m_g, l_g, t_g) in enumerate(zip(mean_grad, l_grad, t_grad)):
            g = 0.0
            if not np.isscalar(m_g):
                g -= 2 * tr_l * np.einsum('ij,ji', c0, m_g)
            if not np.isscalar(l_g):
                ai = t_inv @ t_g
                bi = l_g @ l_inv
                tr_lg = np.trace(l_g)
                g += tr_lg * big_tr + m * np.trace(ai) + n * np.trace(bi)
                g -= tr_l * (np.einsum('ij,ji', ai, c1) + \
                             np.einsum('ij,ji', bi, c2))
                g -= n * m * tr_lg / tr_l 
            grad[i] = g
        return grad

    '''
    -------------------------Prediction method--------------------------------
    '''
    
    def predict(self, x: pd.DataFrame, effects=None):
        """
        Predict data given certain observations.
        
        Uses conditional expectation of the normal distribution method.

        Parameters
        ----------
        x : pd.DataFrame
            DataFrame with missing variables either not present at all, or
            with missing entries set to NaN.
        k : List[EfffectBase], optional
            List of effects to be used for prediction. If None, effects that
            were used during optimization. The default is None.

        Returns
        -------
        None.

        """
        sigma, (m, _) = self.calc_sigma()
        obs = self.vars['observed']
        exo = self.vars['observed_exogenous']
        result = x.copy()
        for v in obs:
            if v not in result:
                result[v] = np.nan
        for v in exo:
            if v not in result:
                if v == '1':
                    result[v] = 1
                else:
                    result[v] = 0
        result = result[obs + exo]
        old_g = self.mx_g.copy()
        self.mx_g = result[exo].values.T
        mean = self.calc_mean(m).reshape((-1, 1), order='F')
        self.mx_g = old_g
        if effects is None:
            effects = deepcopy(self.effects)
        for i, effect in enumerate(effects):
            effect.load(i, self, x, clean_start=False)
        k = [effect.calc_k(self) for effect in effects]
        old_n = self.n_samples
        old_i = self.mx_identity
        self.n_samples = len(x)
        self.mx_identity = np.identity(self.n_samples)
        l = self.calc_l(sigma, k)
        t = self.calc_t(sigma, k)
        l = l / np.trace(l)
        self.n_samples = old_n
        self.mx_identity = old_i
        cov = np.kron(t, l)
        data = result[obs].values.T
        data_shape = data.shape
        data = data.reshape((-1, 1), order='F')
        missing = np.isnan(data).flatten()
        present = ~missing
        cov12 = cov[missing][:, present]
        cov22 = np.linalg.inv(cov[present][:, present])
        mean_m = mean[missing]
        mean_p = mean[present]
        preds = mean_m
        if len(present):
            preds = mean_m + cov12 @ cov22 @ (data[present] - mean_p)
        data[missing] = preds
        result = data.reshape(data_shape, order='F').T
        result = pd.DataFrame(result, columns=obs)
        return result

    def predict_factors(self, x: pd.DataFrame):
        """
        Fast factor estimation method via MAP. Requires complete data.

        Parameters
        ----------
        x : pd.DataFrame
            Complete data of observed variables.

        Returns
        -------
        Factor scores.

        """
        lats = self.vars['latent']
        num_lat = len(lats)
        if num_lat == 0:
            return pd.DataFrame([])
        inners = self.vars['inner']
        obs = self.vars['observed']
        obs_exo = self.vars['observed_exogenous']
        g = []
        for v in obs_exo:
            if v == '1':
                g.append([1] * len(x))
            else:
                g.append(x[v])
        g = np.array(g)
        x = x[obs].values.T
        ks = [effect.calc_k(self) for effect in self.effects]
        ds = self.mxs_d
        lambda_h = self.mx_lambda[:, :num_lat]
        lambda_x = self.mx_lambda[:, num_lat:]
        c = np.linalg.inv(np.identity(self.mx_beta.shape[0]) - self.mx_beta)
        c_1 = c[:num_lat, :]
        c_2 = c[num_lat:, :]
        g1 = self.mx_gamma1; g2 = self.mx_gamma2;
        M_h = x - (g2 + lambda_x @ c_2 @ g1) @ g
        t = lambda_x @ c_2
        L_zh = (t @ self.mx_psi @ t.T + self.mx_theta) * (x.shape[1])
        tr_sigma = np.trace(L_zh) / x.shape[1]
        L_zh += sum(d * np.trace(k) for d, k in zip(ds, ks))
        tr_lzh = np.trace(L_zh)
        try:
            L_zh = chol_inv(L_zh)
        except np.linalg.LinAlgError:
            L_zh = np.linalg.pinv(L_zh)
        T_zh = np.identity(x.shape[1]) * tr_sigma
        T_zh += sum(k * np.trace(d) for d, k in zip(ds, ks))
        try:
            T_zh = chol_inv(T_zh)
        except np.linalg.LinAlgError:
            T_zh = np.linalg.pinv(T_zh)
        t = lambda_h.T @ L_zh
        A = tr_lzh * t @ M_h @ T_zh
        A_0 = tr_lzh * t @ lambda_h
        try:
            L_h = chol_inv(c_1 @ self.mx_psi @ c_1.T)
        except np.linalg.LinAlgError:
            L_h = np.linalg.pinv(c_1 @ self.mx_psi @ c_1.T)
        M = c_1 @ g1 @ g
        A_1 = L_h @ M
        try:
            inv_A0 = chol_inv(A_0)
        except np.linalg.LinAlgError:
            inv_A0 = np.linalg.pinv(A_0)
        A_2 = inv_A0 @ L_h
        A_hat = inv_A0 @ (A + A_1)
        H = solve_sylvester(A_2, T_zh, A_hat)
        return pd.DataFrame(H.T, columns=filter(lambda v: v in lats, inners))

    '''
    -------------------------Best Linear Unbiased Predictor--------------------
    '''
    
    def calc_blup(self, ind_effects=None):
        """
        Estimate random effects values (BLUP).

        Parameters
        ----------
        ind_effects : List[int], optional
            Indices of random effects that will be included to the estimate.
            If None, then all the effects will be estimated and summed up.
            The default is None.

        Returns
        -------
        np.ndarray
        Estimates of random effects.

        """
        if ind_effects is None:
            ind_effects = list(range(len(self.effects)))
        left_effects = list()
        l_i = 0
        t_i = 0
        for i, effect in enumerate(self.effects):
            k = effect.calc_k(self)
            d = self.mxs_d[i]
            if i in ind_effects:
                left_effects.append(np.zeros_like(k))
                t_i += k * np.trace(d)
                l_i += d * np.trace(k)
            else:
                left_effects.append(k)
        sigma, (m, _) = self.calc_sigma()
        z = self.mx_data - self.calc_mean(m)
        t = self.calc_t(sigma, left_effects)
        l = self.calc_l(sigma, left_effects)
        l = l / np.trace(l)
        l_i = l_i / np.trace(l_i)
        t_inv = np.linalg.inv(t)
        t_i_inv = np.linalg.inv(t_i)
        l_i_inv = np.linalg.inv(l_i)
        a = l @ l_i_inv
        b = t_inv @ t_i_inv
        q = z @ t_inv @ t_i
        return solve_sylvester(a, b, q)     
        
        

    '''
    -------------------------Fisher Information Matrix------------------------
    '''

    def calc_fim_ml(self, inverse=False):
        """
        Calculate Fisher Information Matrix.

        Exponential-family distributions are assumed.
        Parameters
        ----------
        inverse : bool, optional
            If True, function also returns inverse of FIM. The default is
            False.

        Returns
        -------
        np.ndarray
            FIM.
        np.ndarray, optional
            FIM^{-1}.

        """
        sigma, (m, c) = self.calc_sigma()
        sigma_grad = self.calc_sigma_grad(m, c)
        mean_grad = self.calc_mean_grad(m, c)
        ks = [effect.calc_k(self) for effect in self.effects]
        k_grad = self.calc_ks_grad()
        t = self.calc_t(sigma, ks)
        l = self.calc_l(sigma, ks)
        try:
            t_inv = chol_inv(t)
        except np.linalg.LinAlgError:
            t_inv= np.linalg.pinv(t)
        try:
            l_inv = chol_inv(l)
        except np.linalg.LinAlgError:
            l_inv = np.linalg.pinv(l)
        l_grad = self.calc_l_grad(sigma, sigma_grad, ks, k_grad)
        t_grad = self.calc_t_grad(sigma, sigma_grad, ks, k_grad)
        tr_l = np.trace(l)
        a = [t_inv @ g if not np.isscalar(g) else None for g in t_grad]
        b = [l_inv @ g if not np.isscalar(g) else None for g in l_grad]
        tr_a = [np.trace(g) if g is not None else g for g in a]
        tr_b = [np.trace(g) if g is not None else g for g in b]
        m_t = [g @ t_inv if not np.isscalar(g) else None for g in mean_grad]
        m_l = [g.T @ l_inv if not np.isscalar(g) else None for g in mean_grad]
        al = [np.trace(g) / tr_l if not np.isscalar(g) else None
              for g in l_grad]
        param_len = len(self.param_vals)
        fim = np.zeros((param_len, param_len))
        n = self.n_samples
        m = self.num_m
        n, m = m, n
        for i in range(param_len):
            for j in range(i, param_len):
                mean = 0
                cov = 0
                mtj = m_t[j]
                mli = m_l[i]
                ai = a[i]
                aj = a[j]
                bi = b[i]
                bj = b[j]
                trai = tr_a[i]
                traj = tr_a[j]
                trbi = tr_b[i]
                trbj = tr_b[j]
                alphai = al[i]
                alphaj = al[j]
                if mli is not None and mtj is not None:
                    mean += tr_l * np.einsum('ij,ji', mli, mtj)
                if ai is not None and aj is not None:
                    cov += n * np.einsum('ij,ji', ai, aj)
                    cov += m * np.einsum('ij,ji', bi, bj)
                    cov += trai * trbj + trbi * traj
                    cov += n * m * alphai * alphaj
                    cov -= n * alphaj * trai + m * alphaj * trbi
                    cov -= n * alphai * traj + m * alphai * trbj
                fim[i, j] = mean + cov / 2
                fim[j, i] = fim[i, j]
        if inverse:
            fim_inv = np.linalg.pinv(fim)
            return (fim, fim_inv)
        return fim
